/*
-- This file is  free  software, which  comes  along  with  SmallEiffel. This
-- software  is  distributed  in the hope that it will be useful, but WITHOUT 
-- ANY  WARRANTY;  without  even  the  implied warranty of MERCHANTABILITY or
-- FITNESS  FOR A PARTICULAR PURPOSE. You can modify it as you want, provided
-- this header is kept unaltered, and a notification of the changes is added.
-- You  are  allowed  to  redistribute  it and sell it, alone or as a part of 
-- another product.
--          Copyright (C) 1994-98 LORIA - UHP - CRIN - INRIA - FRANCE
--            Dominique COLNET and Suzanne COLLIN - colnet@loria.fr 
--                       http://www.loria.fr/SmallEiffel
--
*/

/* This file is automatically included when the Garbage Collector
   is used (default, unless option -no_gc has been selected).
*/

void**stack_bottom;
mch**gcmt=NULL;
int gcmt_max=2048;
int gcmt_used=0;
fsoc*fsocfl=NULL;
rsoc*rsocfl=NULL;
int gc_is_off=1;
unsigned int fsoc_count=0;
unsigned int rsocm_count=0;
unsigned int rso_kb_count=0;
static void*gcmt_tail_addr=NULL;

int na_rounded_size(int s){
  s+=sizeof(rsoh);
  s+=(sizeof(double)-(s&(sizeof(double)-1)));
  return s;
}

void gc_sweep(void){
  mch**gc=gcmt;
  mch**pc=gcmt+1;
  mch**eogcmt=gcmt+gcmt_used;
  rsocfl=NULL;
  if((*gc)->state!=FREE_CHUNK)((*gcmt)->swfp)(*gcmt);
  while (pc<eogcmt) {
    if ((*pc)->state!=FREE_CHUNK)((*pc)->swfp)(*pc);
    if ((*pc)->state==FREE_CHUNK){
      if ((*gc)->state!=FREE_CHUNK) {
	*(gc+1)=*pc;
	gc++;
	pc++;
      }
      else {
	if (((char*)(*gc))+(*gc)->size == ((char*)(*pc))) {
	  ((*gc)->size)+=((*pc)->size);
	  pc++;
	}
	else {
	  (*(gc+1))=*pc;
	  if ((*gc)->type!=FSO_CHUNK) {
	    ((rsoc*)(*gc))->next=rsocfl;
	    rsocfl=((rsoc*)(*gc));
	  }
	  gc++;
	  pc++;
	}
      }
    }
    else {
      *(gc+1)=*pc;
      if (((*gc)->state==FREE_CHUNK)
	  &&((*gc)->type==RSO_CHUNK)) {
	((rsoc*)(*gc))->next=rsocfl;
	rsocfl=((rsoc*)(*gc));
      }
      gc++;
      pc++;
    }
  }
  gcmt_used=(gc-gcmt)+1;
}

void gc_mark(void*p){
  if((p>((void*)*gcmt))&&(p<=gcmt_tail_addr)){
    int i1=0;
    int i2=gcmt_used-1;
    int m=i2>>1;
    mch*c;
    for(;i2>i1;m=((i1+i2)>>1)){
      if(p<=((void*)gcmt[m+1]))i2=m;
      else i1=m+1;}
    c=gcmt[i2];
    if (c->state!=FREE_CHUNK){
      (c->amfp)(c,p);
    }
  }
}

int gc_stack_size(void){
  void*stack_top[2]={NULL,NULL};
  if (stack_top > stack_bottom) {
    return ((void**)stack_top)-((void**)stack_bottom);
  }
  else {
    return ((void**)stack_bottom)-((void**)stack_top);
  }
}

/*
  To delay Garbage Collection when the stack is too large.
  To allow fast increase of ceils.
*/
#define FSOC_LIMIT 1024
#define RSOC_LIMIT 320 

/*
  When stack is too large, collection may be delayed.
*/
#define GCLARGESTACK 50000

int garbage_delayed(void) {
  if (gc_stack_size() > GCLARGESTACK) {
    if (fsoc_count_ceil <= fsoc_count) {
      if (rsocm_count_ceil <= rsocm_count) {
	if ((fsoc_count<FSOC_LIMIT)&&(rsocm_count<RSOC_LIMIT)) {
	  fsoc_count_ceil++;
	  rsocm_count_ceil++;
	return 1;
	}
	else return 0;
      }
      else {
	if (fsoc_count<FSOC_LIMIT) {
	  fsoc_count_ceil++;
	  return 1;
	}
	else return 0;
      }
    }
    else {
      if (rsocm_count_ceil <= rsocm_count) {
	if (rsocm_count<RSOC_LIMIT) {
	  rsocm_count_ceil++;
	  return 1;
	}
	else return 0;
      }
      else return 0;
    }
  }
  else {
    return 0;
  }
}

void gc_update_ceils(void){
  int c;
  /* Compute fsoc_count_ceil :
   */
  if (fsocfl==NULL){
    if (fsoc_count_ceil< FSOC_LIMIT) {
	fsoc_count_ceil<<=1;
    }
    else { 
      c=fsoc_count+(fsoc_count/3);
      if (fsoc_count_ceil < c)
	fsoc_count_ceil=c;
    }
  }
  /* Compute rsocm_count_ceil :
   */
  if (rsocfl==NULL) {
    if (rsocm_count_ceil<RSOC_LIMIT) {
      rsocm_count_ceil<<=1;
    }
    else { 
      c=rsocm_count+(rsocm_count/3);
      if (rsocm_count_ceil < c)
	rsocm_count_ceil=c;
    }
  }
}

static void add_to_gcmt(mch*c,int size){
  mch**p;
  if(gcmt_used==gcmt_max){
    gcmt_max<<=1;
    gcmt=realloc(gcmt,(gcmt_max+1)*sizeof(void*));
  }
  for(p=gcmt+(gcmt_used++ -1);(p>=gcmt)&&(*p>c);p--)
    *(p+1)=*p;
  *(p+1)=c;
  if(((void*)c)>=gcmt_tail_addr){
    gcmt_tail_addr=((char*)c)+size;
  }
}

static fsoc*fsoc_malloc(void){
  mch*r=malloc(FSOC_SIZE);
  fsoc_count++;
  add_to_gcmt((mch*)r,FSOC_SIZE);
  return (fsoc*)r;
}

fsoc*new_fsoc(void){
  fsoc*r;
  if(fsocfl!=NULL){
    r=fsocfl;
    fsocfl=fsocfl->next;
    return r;
  }
  else if(fsoc_count<fsoc_count_ceil){
    return fsoc_malloc();
  }
  else {
    gc_start();
    if(fsocfl!=NULL){
      r=fsocfl;
      fsocfl=fsocfl->next;
      return r;
    }
    else {
      return fsoc_malloc();
    }
  }
}

static rsoh*allocate_rso_from_store(na_env*nae,int size){
  rsoh*r=(nae->store);
  int store_left=r->header.size-size;
  if ((store_left)>sizeof(rsoh)){
    r->header.size=size;
    ((char*)(nae->store))+=size;
    nae->store->header.size=store_left;
    nae->store->header.magic_flag=RSOH_FREE;
  }
  else {
    nae->store=NULL;
  }
  (r->header.magic_flag)=RSOH_UNMARKED;
  return r;
}

void rsoh_before_sweep(na_env*nae){
  nae->chunk_list=NULL;
  nae->store=NULL;
  nae->store_chunk=NULL;
}

static void rsoc_sweep(rsoc*c){
  na_env*nae=c->nae;
  rsoh*gp=(rsoh*)&(c->padding);
  rsoh*pp;
  rsoh*eoc=((rsoh*)(((char*)c)+c->header.size));
  c->free_list_of_large=NULL;
  if (c->header.size > RSOC_SIZE) {
    if (gp->header.magic_flag == RSOH_MARKED) {
      gp->header.magic_flag=RSOH_UNMARKED;
      c->next=nae->chunk_list;
      nae->chunk_list=c;
    }
    else {
      c->header.state=FREE_CHUNK;
    }
    return;
  }
  while (gp<eoc) {
    while (gp->header.magic_flag == RSOH_MARKED) {
      gp->header.magic_flag=RSOH_UNMARKED;
      ((char*)gp)+=gp->header.size;
      if(gp>=eoc) {
	c->next=nae->chunk_list;
	nae->chunk_list=c;
	return;
      }
    }
    gp->header.magic_flag=RSOH_FREE;
    pp=(rsoh*)(((char*)gp)+gp->header.size);
    while ((pp<eoc)&&(pp->header.magic_flag != RSOH_MARKED)) {
      pp->header.magic_flag=RSOH_FREE;
      gp->header.size+=pp->header.size;
      ((char*)pp)+=pp->header.size;
    }
    if (gp->header.size >= RSOC_MIN_STORE) {
      if (nae->store==NULL) {
	nae->store=gp;
	nae->store_chunk=c;
      }
      else if (nae->store->header.size < gp->header.size) {
	((fll_rsoh*)nae->store)->next=nae->store_chunk->free_list_of_large;
	nae->store_chunk->free_list_of_large=((fll_rsoh*)nae->store);
	nae->store=gp;
	nae->store_chunk=c;
      }
      else {
	((fll_rsoh*)gp)->next=c->free_list_of_large;
	c->free_list_of_large=((fll_rsoh*)gp);
      }
    }
    gp=pp;
  }
  if (((rsoh*)(&c->padding))->header.size >= 
      (c->header.size-sizeof(rsoc)+sizeof(rsoh))){
    c->header.state=FREE_CHUNK;
    nae->store_chunk=NULL;
    nae->store=NULL; 
  }
  else{
    c->next=nae->chunk_list;
    nae->chunk_list=c;
  }
}

static rsoc MRSOC = {{0,USED_CHUNK,RSO_CHUNK,
		      (void(*)(mch*,void*))gcna_align_mark,
		      (void(*)(mch*))rsoc_sweep},
		     NULL, NULL, NULL, {0,0}};

static void rsoc_malloc(na_env*nae,int size){
  rsoc*r=malloc(size);
  rsocm_count++;
  *r=MRSOC;
  r->header.size=size;
  r->nae=nae;
  nae->store=(&(r->padding));
  nae->store->header.size=size-sizeof(rsoc)+sizeof(rsoh);
  nae->store_chunk=r;
  r->next=nae->chunk_list;
  nae->chunk_list=r;

  add_to_gcmt((mch*)r,size);
}

static rsoc*rsocfl_first_fit(int size) {
  rsoc*pc=NULL;
  rsoc*c=rsocfl;
  while (c != NULL) {
    if (c->header.size>=size) {
      if (pc==NULL) {
	rsocfl = c->next;
      }
      else {
	pc->next=c->next;
      }
      return c;
    }
    pc=c;
    c=c->next;
  }
  return NULL;
}

static rsoc* rsocfl_best_fit(int size) {
  int best_size;
  rsoc *pc,*best_pc,*best_c, *c;
  pc=NULL;
  best_pc=NULL;
  best_c=NULL;
  c=rsocfl;
  while ((NULL!=c)&&(NULL==best_c)){
    if (c->header.size>=size){
      best_c=c;
      best_pc=pc;
      best_size=c->header.size;
    }
    pc=c;
    c=c->next;
  }
  if (NULL==c)
    return NULL;
  do{
    if ((c->header.size>=size)&&(c->header.size<best_size)){	
      best_c=c;
      best_pc=pc;
      best_size=c->header.size;
    }
    pc=c;
    c=c->next;
  } while(c!=NULL);
  if (NULL==best_pc) {
    rsocfl = best_c->next;
  }
  else {
    best_pc->next=best_c->next;
  }
  return best_c;
}

static int get_store_in(rsoc*c,int size){
  na_env*nae=c->nae;
  fll_rsoh*pf=NULL;
  fll_rsoh*f=c->free_list_of_large;
  while (f != NULL) {
    if (f->rsoh_field.header.size>=size) {
      nae->store=(rsoh*)f;
      nae->store_chunk=c;
      if (pf == NULL) {
	c->free_list_of_large = f->next;
      }
      else {
	pf->next = f->next;
      }
      return 1;
    }
    pf = f;
    f = f->next;
  }
  return 0;
}

static int get_store(na_env*nae,int size) {
  if (nae->store!=NULL && nae->store->header.size >= RSOC_MIN_STORE) {
    ((fll_rsoh*)nae->store)->next=nae->store_chunk->free_list_of_large;
    nae->store_chunk->free_list_of_large=((fll_rsoh*)nae->store);
    nae->store=NULL;
  }
  if ((nae->store_chunk!=NULL)&&
      (get_store_in(nae->store_chunk,size))) {
    return 1;
  }
  else {
    rsoc*c=nae->chunk_list;
    while (c != NULL) {
      if (get_store_in(c,size))
	return 1;
      c = c->next;
    }
    size+=(sizeof(rsoc)-sizeof(rsoh));
    c=rsocfl_best_fit(size);
    if (NULL!=c){
      if ((c->header.size > RSOC_SIZE)&&
	  (c->header.size-size > RSOC_MIN_STORE*4)) {
	int left = c->header.size - size;
	c->header.size=left;
	c->next=rsocfl;
	rsocfl=c;
	c=(rsoc*)(((char*)c)+left);
	add_to_gcmt((mch*)c,size);
      }
      else {
	size=c->header.size;
      }
      *c=MRSOC;
      c->header.size=size;
      c->nae=nae;
      nae->store=(&(c->padding));
      nae->store->header.size=size-sizeof(rsoc)+sizeof(rsoh);
      nae->store_chunk=c;
      c->next=nae->chunk_list;
      nae->chunk_list=c;
      return 1;
    }
  }
  return 0;
}

char*new_na(na_env*nae,int size){ 
  rsoh*r;
  size=na_rounded_size(size);
 
  if ((NULL!=nae->store)&&(size<=(nae->store->header.size))){
    r=allocate_rso_from_store(nae,size);
  }
  else if (get_store(nae,size)) {
    r=allocate_rso_from_store(nae,size);
  }
  else if(rsocm_count<rsocm_count_ceil){
    if((size+sizeof(rsoc))>RSOC_SIZE){
      rsoc_malloc(nae,size+sizeof(rsoc));/*...-sizeof(rsoh)?*/
      r=allocate_rso_from_store(nae,size);
    }
    else{
      rsoc_malloc(nae,RSOC_SIZE);
      r=allocate_rso_from_store(nae,size);
    }
  }
  else{
    gc_start();
    if ((NULL!=nae->store)&&(size<=(nae->store->header.size))){
      r=allocate_rso_from_store(nae,size);
    }
    else if (get_store(nae,size)) {
      r=allocate_rso_from_store(nae,size);
    }
    else {
      if((size+sizeof(rsoc))>RSOC_SIZE){
	rsoc_malloc(nae,size+sizeof(rsoc));
	r=allocate_rso_from_store(nae,size);
      }
      else {
	rsoc_malloc(nae,RSOC_SIZE);
	r=allocate_rso_from_store(nae,size);
      }
    }
  }
  ((void)memset((r+1),0,r->header.size-sizeof(rsoh)));
  return (char*)(r+1);
}

void gcna_align_mark(rsoc*c,void*o){
  na_env*nae=c->nae;
  fll_rsoh*f,*pf;
  char*b=(char*)&(c->padding);
  if(((char*)o)>(((char*)c)+c->header.size))
    return;
  if(((((char*)o)-((char*)c))%sizeof(double))!=0)
    return;
  if((((rsoh*)o)-1)->header.magic_flag!=RSOH_UNMARKED)
    return;
  if(((char*)o)<((char*)(c+1)))
    return;
  if(c->header.size>RSOC_SIZE){
    if (o==(c+1))
      nae->gc_mark(o);
    return;
  }
  pf=NULL;
  f=c->free_list_of_large;
  while ((f!=NULL)&&(f<((fll_rsoh*)o))){
    pf=f;
    f=f->next;
  }
  if (pf==NULL)
    pf=(fll_rsoh*)b;
  while ((((rsoh*)pf)+1)<(rsoh*)o)
    ((char*)pf)+=pf->rsoh_field.header.size;
  if ((((rsoh*)pf)+1)==(rsoh*)o)
    nae->gc_mark(o);
}

void rsocfl_count(void) {
  int r=0;
  int max=0;
  rsoc*p=rsocfl;
  while (p!=NULL) {
    r++;
    p=p->next;
  }
  printf(" rsocfl count = %d\n",r);
  if (r>0) {
    printf(" rsocfl = [",r);
    p=rsocfl;
    while (p!=NULL) {
      printf("%d",p->header.size);
      if(max < p->header.size) max=p->header.size;
      p=p->next;
      if (p!=NULL)printf(",");
    }
    printf("]\n");
    printf(" rsocfl max = %d\n",max);
  }
}

int fsocfl_count(void) {
  int r=0;
  fsoc*p=fsocfl;
  while (p!=NULL) {
    r++;
    p=p->next;
  }
  return r;
}
